"""
This file contains functions useful for solving differential equationss 
which occur commonly in a 1st semster differential equations course.

* desolve -- Computes the "general solution" to a 1st or 2nd order 
             ODE via maxima.

* desolve_laplace -- Solves an ODE using laplace transforms via maxima. 
                     Initials conditions are optional.

* desolve_system -- Solves any size system of 1st order odes using maxima. 
                    Initials conditions are optional.

AUTHOR: David Joyner (3-2006)

Other functions for solving DEs are given in functions/elementary.py.

"""

##########################################################################
#  Copyright (C) 2006 David Joyner <wdjoyner@gmail.com> and William Stein
#
#  Distributed under the terms of the GNU General Public License (GPL):
#
#                  http://www.gnu.org/licenses/
##########################################################################

def desolve(de,vars):
    """
    Solves a 1st or 2nd order ODE via maxima. Initials conditions
    are not given.

    INPUT: 
        de    -- a string representing the ODE 
                 (eg, de = "diff(f(x),x,2)=diff(f(x),x)+sin(x)")
        vars  -- a list of strings representing the variables
                 (eg, vars = ["x","y"], if x is the independent
                  variable and y is the dependent variable)

    EXAMPLES:
        sage: desolve("x^2*diff(y,x) + 3*y*x = sin(x)/x",["x","y"])
        (%c - cos(x))/x^3

    AUTHOR: David Joyner (1-2006)
    """
    maxima("depends("+vars[1]+","+vars[0]+");")
    maxima("de:"+de+";")
    cmd = "ode2("+de+","+vars[1]+","+vars[0]+");"
    print maxima.eval(cmd)
    #maxima.eval(cmd)

def desolve_laplace(de,vars,ics=None):
    """
    Solves an ODE using laplace transforms via maxima. Initials conditions
    are optional.

    INPUT: 
        de    -- a string representing the ODE 
                 (eg, de = "diff(f(x),x,2)=diff(f(x),x)+sin(x)")
        vars  -- a list of strings representing the variables
                 (eg, vars = ["x","f"], if x is the independent
                  variable and f is the dependent variable)
        ics   -- a list of numbers representing initial conditions,
                 with symbols allowed which are represented by strings
                 (eg, f(0)=1, f'(0)=2 is ics = [0,1,2])
           
    EXAMPLES:
        sage: desolve_laplace("diff(f(x),x,2)=2*diff(f(x),x)-f(x)",["x","f"])
        x*%e^x*(at('diff(f(x),x,1),x = 0)) - f(0)*x*%e^x + f(0)*%e^x
        sage: desolve_laplace("diff(f(x),x,2)=2*diff(f(x),x)-f(x)",["x","f"],[0,1,2])
        x*%e^x + %e^x

    WARNING: 
        The second SAGE command in the above example sets the values of f(0) and f'(0) in maxima, so
        subsequent ODEs involving these variables will have these initial conditions
        automatically imposed. 

    AUTHOR: David Joyner (1-2006)
    """
    maxima("de:"+de+";")
    if ics!=None:
        d = len(ics)
        for i in range(0,d-1):
            ic = "atvalue(diff("+vars[1]+"("+vars[0]+"),"+str(vars[0])+","+str(i)+"),"+str(vars[0])+"="+str(ics[0])+","+str(ics[1+i])+")"
            maxima(ic)
            #print i,ic
    cmd = "desolve("+de+","+vars[1]+"("+vars[0]+"));"
    print maxima.eval(cmd)
    #maxima.eval(cmd)

def desolve_system(des,vars,ics=None):
    """
    Solves any size system of 1st order odes using maxima. Initials conditions
    are optional.


    INPUT: 
        de    -- a list of strings representing the ODEs in maxima notation
                 (eg, de = "diff(f(x),x,2)=diff(f(x),x)+sin(x)")
        vars  -- a list of strings representing the variables
                 (eg, vars = ["t","x","y"], where t is the indeendent variable
                  and x,y the dependent variables)
        ics   -- a list of numbers representing initial conditions
                 (eg, x(0)=1, y(0)=2 is ics = [0,1,2])

    WARNING: 
        The given ics sets the initial values of the dependent vars in maxima, so
        subsequent ODEs involving these variables will have these initial conditions
        automatically imposed. 

    EXAMPLES:
        sage: des=["'diff(x(t),t)=-4*y(t)","'diff(y(t),t)=-x(t)"]
        sage: vars = ["t","x","y"]
        sage: desolve_system(des,vars)
        [x(t) = (2*y(0) + x(0))*%e^-(2*t)/2 - (2*y(0) - x(0))*%e^(2*t)/2,y(t) = (2*y(0) - x(0))*%e^(2*t)/4 + (2*y(0) + x(0))*%e^-(2*t)/4]
	sage: des=["'diff(x(t),t)=-4*y(t)","'diff(y(t),t)=-x(t)"]
    	sage: vars = ["t","x","y"]
    	sage: desolve_system(des,vars,[0,270,90])
        [x(t) = 45*%e^(2*t) + 225*%e^-(2*t),y(t) = 225*%e^-(2*t)/2 - 45*%e^(2*t)/2]
    	sage: des=["'diff(x(t),t)=-4*y(t)","'diff(y(t),t)=-z(t)","'diff(z(t),t)=-2*x(t)"]
    	sage: vars = ["t","x","y","z"]
    	sage: ics=[0,270,90,100]
    	sage: desolve_system(des,vars,ics)
    	[x(t) = %e^t*(260*cos(sqrt(3)*t)/3 - 80*sin(sqrt(3)*t)/sqrt(3)) + 550*%e^ - (2*t)/3,y(t) = %e^t*(85*sin(sqrt(3)*t)/sqrt(3) - 5*cos(sqrt(3)*t)/3) + 275*%e^ - (2*t)/3,z(t) = %e^t*( - 90*sin(sqrt(3)*t)/sqrt(3) - 250*cos(sqrt(3)*t)/3) + 550*%e^ - (2*t)/3]
 
    AUTHOR: David Joyner (3-2006)
    """
    d = len(des)
    for i in range(d):
        cmd="de:"+des[int(i)]+";"
        maxima(cmd)
    desstr = "["+",".join(des)+"]"
    #print desstr
    d = len(vars)
    varss = list(vars[i]+"("+vars[0]+")" for i in range(1,d))
    varstr = "["+",".join(varss)+"]"
    #print varstr
    if ics!=None:
        d = len(ics)
        for i in range(1,d):
            ic = "atvalue("+vars[int(i)]+"("+vars[0]+"),"+str(vars[0])+"="+str(ics[0])+","+str(ics[int(i)])+")"
            maxima(ic)
    cmd = "desolve("+desstr+","+varstr+");"
    #print cmd
    print maxima.eval(cmd)
    #maxima.eval(cmd)