r"""

This file contains a few routines which are useful to teaching a lower
level ODE or a calculus II or III course.

* direction_field_plot -- plots the direction field of y' = f(x,y), y(a)=c,
                          and sketches an approximate solution.

* direction_field_with_isocline_plot - same as direction_field_plot but
                                       also includes some isoclines.

* vector_field_plot -- plots the vector field vf = [f,g] (f,g scalar-valued 
                       functions in 2 variables)

AUTHOR: David Joyner (9-2006)
     
"""


#*****************************************************************************
#       Copyright (C) 2006 David Joyner <wdj@usna.edu>
#                     2006 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

def direction_field_plot(f,a,c,h,xrange,yrange):
    """
    This plots the direction field in the rectangle 
    (xrange[0],xrange[1]) x (yrange[0],yrange[1])
    and plots using Euler's method the numerical solution of 
    the 1st order ODE y' = f(x,y), y(a)=c. We assume (a,c) is in this
    rectangle.
    EXAMPLES:
        sage: x,y=PolynomialRing(QQ,2).gens()
        sage: direction_field_plot(5*x+y-5,0,1,1/10,[-1,1],[-5,5])
        [[0, 1],
         [1/10, 3/5],
         [1/5, 21/100],
         [3/10, -169/1000],
         [2/5, -5359/10000],
         [1/2, -88949/100000],
         [3/5, -1228439/1000000],
         [7/10, -15512829/10000000],
         [4/5, -185641119/100000000],
         [9/10, -2142052309/1000000000],
         [1, -24062575399/10000000000]]
    Your browser will open and the plot of the direction field (in black) with the 
    approximate solution (in blue) will appear. Returned will be a list of
    points on the graph of the approximate solution.
    """
    X1 = max(xrange)
    x1 = min(xrange)
    Y1 = max(yrange)
    y1 = min(yrange)
    x0 = a
    y0 = c
    n = int((1.0)*(X1-x0)/h)
    x00 = x0; y00 = y0
    soln = [[a,c]]
    for i in range(n+1):
        y00 = y00+h*f(x00,y00)
        x00=x00+h
        soln.append([x00,y00])
    M = int((1.0)*(X1-x1)/h)
    N = int((1.0)*(Y1-y1)/h)
    pts = [[x1+h*(X1-x1)*i,y1+h*(Y1-y1)*j] for i in range(M) for j in range(N)]
    #print pts
    P = add([line([[p[0],p[1]],[p[0]+h/2,p[1]+(h/2)*f(p[0],p[1])]], rgbcolor=(1/4,1/8,3/4)) for p in pts])
    Q = line(soln, rgbcolor=(1/4,1/8,3/4))
    df = P+Q
    df.show(xmin=xrange[0], xmax=xrange[1], ymin=yrange[0], ymax=yrange[1])
    return soln[:-1]


def direction_field_with_isocline_plot(f,a,c,h,xrange,yrange):
    """
    This plots the direction field and some isoclines in the rectangle 
    (xrange[0],xrange[1]) x (yrange[0],yrange[1])
    and plots using Euler's method the numerical solution of 
    the 1st order ODE y' = f(x,y), y(a)=c. We assume (a,c) is in this
    rectangle. The loop for computing the points on the isoclines
    is very slow. This is because SAGE has no implicit plot method 
    using matplotlib yet.

    EXAMPLES:
        sage: x,y=PolynomialRing(QQ,2).gens()
        sage: direction_field_with_isocline_plot(5*x+y-5,0,1,1/10,[-1,1],[-5,5])
        [[0, 1],
         [1/10, 3/5],
         [1/5, 21/100],
         [3/10, -169/1000],
         [2/5, -5359/10000],
         [1/2, -88949/100000],
         [3/5, -1228439/1000000],
         [7/10, -15512829/10000000],
         [4/5, -185641119/100000000],
         [9/10, -2142052309/1000000000],
         [1, -24062575399/10000000000]]
    Your browser will open and the plot of the direction field (in black) with the 
    approximate solution (in blue) will appear. There will be 3 green
    isoclines, one of which passes through the initial condition (a,c)=(0,1).
    Returned will be a list of points on the graph of the approximate solution.
    """
    X1 = max(xrange)
    x1 = min(xrange)
    Y1 = max(yrange)
    y1 = min(yrange)
    x0 = a
    y0 = c
    n = int((1.0)*(X1-x0)/h)
    x00 = x0; y00 = y0
    soln = [[a,c]]
    for i in range(n+1):
        y00 = y00+h*f(x00,y00)
        x00=x00+h
        soln.append([x00,y00])
    M = int((1.0)*(X1-x1)/h)
    N = int((1.0)*(Y1-y1)/h)
    pts = [[x1+h*(X1-x1)*i,y1+h*(Y1-y1)*j] for i in range(M) for j in range(N)]
    #print pts
    P = add([line([[p[0],p[1]],[p[0]+h/2,p[1]+(h/2)*f(p[0],p[1])]], rgbcolor=(1/4,1/8,3/4)) for p in pts])
    Q = line(soln, rgbcolor=(1/4,1/8,3/4))
    isocline_pts1 = []
    isocline_pts2 = []
    isocline_pts3 = []
    for i in range(100):
        for j in range(100):
            xx = x1 + (X1-x1)*i/100
            yy = y1 + (Y1-y1)*j/100
            #print xx,yy,abs(f(xx,yy)-f(a,c))
            if abs(f(xx,yy)-f(a,c))<1/5:
               isocline_pts1.append([xx,yy])
            if abs(f(xx,yy)-f(a,c)-1)<1/5:
               isocline_pts2.append([xx,yy])
            if abs(f(xx,yy)-f(a,c)+1)<1/5:
               isocline_pts3.append([xx,yy])
    #R = add([circle((p[0],p[1]), 1/2, rgbcolor=(0,1,0)) for p in isocline_pts])
    R1 = line(isocline_pts1, rgbcolor=(0,1,0), thickness=0.1)
    R2 = line(isocline_pts2, rgbcolor=(0,1,0), thickness=0.1)
    R3 = line(isocline_pts3, rgbcolor=(0,1,0), thickness=0.1)
    #R.show()
    df = P+Q+R1+R2+R3
    df.show(xmin=xrange[0], xmax=xrange[1], ymin=yrange[0], ymax=yrange[1])
    return soln[:-1]


def vector_field_plot(vf,xrange,yrange):
    """
    This plots the vector field vf = [f,g] in the rectangle 
    (xrange[0],xrange[1]) x (yrange[0],yrange[1]).
    EXAMPLES:
        sage: x,y=PolynomialRing(QQ,2).gens()
        sage: vector_field_plot([y,-x],[-1,1],[-1,1])
        'Vector field plot for [x1, -1*x0]'
    Your browser will open and the plot of the vector field (in green) will appear. 
    The string title for the plot (as above) is returned.
    """
    h = 1/20
    f = vf[0]
    g = vf[1]
    X1 = max(xrange)
    x1 = min(xrange)
    Y1 = max(yrange)
    y1 = min(yrange)
    M = int((1.0)*(X1-x1)/h)
    N = int((1.0)*(Y1-y1)/h)
    pts = [[x1+h*(X1-x1)*i,y1+h*(Y1-y1)*j] for i in range(M) for j in range(N)]
    #print pts
    #p0 = pts[0]
    #P = line([p0,[p0[0]+h*f(p0[0],p0[1]),p0[1]+h*g(p0[0],p0[1])]], rgbcolor=(1,0.5,0.2))
    #for p in pts:
    #    P = P + line([p,[p[0]+(h/2)*f(p[0],p[1]),p[1]+(h/2)*g(p[0],p[1])]], rgbcolor=(1,0.5,0.2))
    P = add([line([p,[p[0]+h*f(p),p[1]+h*g(p)]], rgbcolor=(0.5,0.5,0.2)) for p in pts])
    P.show(xmin=xrange[0], xmax=xrange[1], ymin=yrange[0], ymax=yrange[1])
    return "Vector field plot for %s"%vf