"""
Newton-Raphson which works with lambda functions.
Stops when the approximation is within eps of the
desired root.
 
EXAMPLES:
    sage: attach 'newton_raphson.sage'
    sage: f = lambda x : -49*exp(x/2) + 251*exp(-x/2) - 1
    sage: newton_raphson(f,1.6,0.000001,0.00001)
    1.6246136830923981
"""
def newton_raphson(f, x, dx, eps):
    df = lambda x : (f(x+dx)-f(x))/dx
    i = 0
    while i<1000:
        i=i+1
        x1 = x-f(x)/df(x)
        t = abs(x1-x)
        if t < eps: break
        x = x1
    return x

"""
Newton-Raphson which works with lambda functions.
Stops when the number of steps reaches N.
 
EXAMPLES:
    sage: attach 'newton_raphson.sage'
    sage: f = lambda x : -49*exp(x/2) + 251*exp(-x/2) - 1
    sage: newton_raphson2(f,1.6,0.000001,10)
    1.6246156086675705
"""
def newton_raphson2(f, x, dx, N):
    df = lambda x : (f(x+dx)-f(x))/dx
    for i in range(N):
        x1 = x-f(x)/df(x)
        x = x1
    return x

"""
This version prints out the sequence of successive approximations, 
suitable for plotting.

EXAMPLES:
    sage: f = lambda x : x^3 - 3*x^2 -1
    sage: L = newton_raphson3(f,2.5,0.0000001,5)
    sage: L
    [[2.5000000000000000, -4.1250000000000000],
     [3.5999998725110798, 6.7759977969915894],
     [3.2078703415436505, 1.1390756411114147],
     [3.1098775425845280, 0.062662889189841309],
     [3.1038259346814483, 0.00023158578437332267]]
    sage: P = add([circle(L[i],0.02) for i in range(5)])
    sage: Q = add([line([L[i],L[i+1]],rgbcolor=(i/(i+2),cos(1/(i+1)),1/(i+1))) for i in range(4)])

Now type show(P+Q) to see the successive approximations (as black points) and
lines segments between them.
"""
def newton_raphson3(f, x, dx, N):
    pts = []
    df = lambda x : (f(x+dx)-f(x))/dx
    for i in range(N):
        pts.append([x,f(x)])
        x1 = x-f(x)/df(x)
        x = x1
    return pts
